;
;                                     DMC
;                             DDKEY Macro Compiler
;
;  Source code for assembly under WASM, the Wolfware Assembler, by Eric Tauck.
;
; Author:  Richard Cowell
; Address: S.K. Heights #6
;          Suwa 239, Takatsu-ku
;          Kawasaki, Kanagawa  213
;          JAPAN
;

     JUMP+
     UNUSED+
          mov       sp,STACKTOP
          call      Init                ; initialization procedure

          test      Flag,MACINPUT       ; is input file a macro file?
          jz        MakeMac             ;   no, jump

          call      MacToText           ; convert
          jmps      Output

MakeMac
          call      TextToMac           ; convert

Output
          call      WriteBuff           ; write output buffer to file or screen
          mov       ax,04C00h           ; terminate program
          int       21h


; PROCEDURES 
;͸
;                 Init                 
; checks for option                    
; opens input file                     
; reads input file into buffer         
; gets rows/screen                     
;                                      
; Out: bp = ptr to filename            
;      cx = bytes read into buffer     
;      MACINPUT flag set if input file 
;        is a macro file               
;;
Init      PROC      NEAR
;=======================================
;[I. Command Line Arguments ]

          call      GetArg              ; any command line arguments?
          jnc       Go_ini              ;   yes, si = ptr to 1st argument, jump
U_ini
          mov       si,offset Umsg      ; show usage message and terminate
          call      ErrorExit

     ;[ check for help wanted and option ]
Go_ini
          cmp       byte [si],'?'       ; is first byte '?' ?
          jz        U_ini               ;   yes, jump to show usage message
          cmp       byte [si],'/'       ; option?
          jnz       Open_ini            ;   no, jump to open file
          and       byte [si+1],05Fh    ; capitalize option letter
          cmp       byte [si+1],'N'     ; valid option?
          jnz       U_ini               ;   no, jump to show usage message
          or        Flag,NEWKEY         ;   yes, set flag
          call      GetArg              ; get next argument
          jc        U_ini               ; if none, error. jump to show usage

;=======================================
;[II. Open File ]
Open_ini
          mov       bp,si               ; out: bp = ptr to input filename
          mov       dx,si               ; dx = ptr to filename
          mov       ax,03D00h           ; open file, read access (al = 0)
          int       21h
          jc        Xo_ini              ; jump on error
          mov       bx,ax               ; bx = handle

;=======================================
;[III. Read File ]
     ;[ read first 5 bytes to determine if it's a text file ]
          mov       dx,TEXTBUFF         ; dx = buffer address
          mov       cx,5                ; cx = bytes to read
          mov       ah,03Fh             ; read file  (bx = handle)
          int       21h
          jc        Xr_ini              ; jump on error
          cmp       ax,cx               ; requested bytes read?
          jnz       Xr_ini              ;   no, error. jump

     ;[ check for 'DDKEY' file ID ]
          cmp       word [TEXTBUFF],'DD'     ; are first 2 bytes 'DD'?
          jnz       Mac_ini                  ;   no, must be macro file, jump
          cmp       word [TEXTBUFF+2],'KE'   ; if first 2 bytes are 'DD' but
          jnz       Xi_ini                   ;   'KEY' does not follow, it's
          cmp       byte [TEXTBUFF+4],'Y'    ;    not an acceptable file.
          jnz       Xi_ini                   ;    error. jump

     ;[ text file ]
          mov       si,pArg             ; pArg = ptr to command line
          cmp       byte [si],0         ; output filename for macros exists?
          jz        Xn_ini              ;   no, error. jump
          mov       dx,TEXTBUFF+5       ; address of buffer
          mov       cx,TEXTSIZE-5       ; bytes available
          jmps      Read_ini            ; jump to read

     ;[ macro file ]
Mac_ini
          or        Flag,MACINPUT       ; set flag
          mov       si,TEXTBUFF         ; copy 5 bytes from TEXTBUFF to MACBUFF
          mov       di,MACBUFF
          mov       cx,5
          rep
          movsb
          mov       dx,di               ; address of buffer
          mov       cx,MACSIZE-5        ; bytes available

     ;[ read file into buffer ]
Read_ini
          mov       ah,03Fh             ; read file  (bx = handle)
          int       21h
          jc        Xr_ini              ; jump on error
          mov       cx,ax               ; cx = bytes read
          jcxz      Xr_ini              ; if no bytes read, error. jump
          add       cx,5                ; include initial 5 bytes in total

          cmp       cx,TEXTSIZE         ; (For macro file, cx << TEXTSIZE. So,
                                        ;  ZF clear.) If cx = TEXTSIZE, then
                                        ;  the text buffer is completely full.
          jz        Xf_ini              ;  If so, file is too large. jump

     ;[ close file ]
          mov       ah,03Eh             ; (bx = handle)
          int       21h

;=======================================
;[IV. Get Rows/Screen ]
          mov       ax,040h             ; BIOS Data Area
          mov       es,ax
          seg       es
          mov       al,[084h]           ; rows/screen - 1
          mov       Rows,al
          mov       ErrCount,al
          mov       ax,cs
          mov       es,ax
          ret
;---------------------------------------
     ;[ error handling ]
Xo_ini
          mov       si,offset Xomsg     ; "can't open" message
          jmps      XX_ini
Xi_ini
          mov       si,offset Ximsg     ; "bad input file" message
          jmps      X_ini
Xn_ini
          mov       si,offset Xnmsg     ; "need output file" message
          jmps      X_ini
Xr_ini
          mov       si,offset Xrmsg     ; "read problem" message
          jmps      X_ini
Xf_ini
          mov       si,offset Xfmsg     ; "file too big" message
X_ini
          mov       ah,03Eh             ; close file  (bx = handle)
          int       21h
XX_ini
          call      ErrorExit           ; write message and terminate program
          ENDP
;͸
;              GetArg                  
;    Scans command line arguments      
;    returning pointers to ASCIZ       
;    strings successively on each call 
;                                      
; Success:  si = ptr to argument       
;           CF clear                   
; Failure:  CF set                     
;;
GetArg    PROC      NEAR
          mov       si,pArg             ; Initially, pArg = 80h (See DATA)
          cmp       byte [si],0         ; any arguments?
          jz        None_arg            ;   no, done
          cmp       si,080h             ; first run?
          jnz       B_arg               ;    no, jump
A_arg                                   ;    yes, skip 1st byte = length
          inc       si
B_arg
          cmp       byte [si],32        ; skip spaces
          jz        A_arg
          cmp       byte [si],9         ; skip tabs
          jz        A_arg
          push      si                  ; save pointer to current arg

C_arg
          inc       si                  ; look for end
          cmp       byte [si],' '       ; ASCII value > space?
          ja        C_arg               ;    yes, jump to advance
          cmp       byte [si],CR        ; if it's a CR...
          mov       byte [si],0         ;    (zero out end)
          jz        D_arg               ;               ...leave pointer at '0'
          inc       si                  ; else move past '0' to signal args exist
D_arg
          mov       pArg,si             ; save this position on command line
          pop       si                  ; return pointer to arg
          clc
          ret
;---------------------------------------
None_arg
          stc
          ret
          ENDP
; MACRO-TO-TEXT 
;͸
;             MacToText                
;      converts macros to text         
;                                      
;  In: MACBUFF contains macro file     
;      MACINPUT flag set               
;      bp = ptr to input filename      
; Out: di = limit of text in TEXTBUFF  
;;
MacToText PROC      NEAR
;=======================================
;[ I. Line 1 ]

     ;[ write 'DDKEY' to beginning of TEXTBUFF ]
          mov       di,TEXTBUFF         ; di = ptr to TEXTBUFF throughout this
          mov       si,offset Ddkey     ;   procedure, and it is preserved in
          mov       cx,DDKEYLEN         ;   all subroutines
          rep
          movsb

     ;[ write filename ]
          mov       si,bp               ; si = ptr to input filename
Fil_mtt
          movsb
          cmp       byte [si],0
          jnz       Fil_mtt

     ;[ write CRLF twice ]
          mov       ax,0A0Dh
          stosw
          stosw

;=======================================
;[ II. Line 2 ]

     ;[ write bytes/mac ]
          mov       si,MACBUFF
          lodsw                         ; al = macro length, ah = 0
          cmp       ax,5                ; check bytes to make sure this is
          jb        Xi_mtt              ;   a macro file. if not in range, jump
          cmp       ax,96
          ja        Xi_mtt
          mov       bp,si               ; bp = start of first macro
          mov       dx,ax               ; dx = macro length
          sub       dx,2                ; dx = macro length - key length
          mov       BytPerMac,dx        ; store bytes available for definition
          call      WriteNum            ; write macro length to TEXTBUFF

     ;[ write ' == Number...' ]
          mov       si,offset Nmsg
          mov       cx,NLEN
          rep
          movsb

;=======================================
;[ III. Line 3+ ]

     ;[ write macros ]
Next_mtt
          mov       si,bp               ; si = ptr to start of macro = key
          lodsw                         ; key: ah = scan, al = ASCII
          or        ax,ax               ; end of macro list?
          jz        Done_mtt            ;    yes, jump

          cmp       di,TBUFFLIM         ; near the end of TEXTBUFF?
          ja        Xl_mtt              ;   yes, can't handle any more. jump
          mov       bp,si               ; save start of definition in bp
          call      WriteKey            ; write key symbol to TEXTBUFF
          call      WriteDef            ; write definition symbols to TEXTBUFF
          add       bp,BytPerMac        ; bp = ptr to start of next macro
          jmps      Next_mtt            ; jump to process

Done_mtt
          ret
;---------------------------------------
     ;[ error handling ]
Xi_mtt
          mov       si,offset Ximsg     ; "bad input file" message
          jmps      X_mtt
Xl_mtt
          mov       si,offset Xlmsg     ; "text file too long" message
X_mtt
          call      ErrorExit
          ENDP
;͸
;              WriteKey                
; writes symbol for key and equal sign 
; to TEXTBUFF                          
;                                      
;  In: ah = scan or trigger code       
;      al = ASCII code                 
;      di = destination in TEXTBUFF    
; Out: di updated                      
;;
WriteKey  PROC      NEAR
          or        al,al               ; ASCII code = 0?
          jz        Spec_wk             ;   yes, jump

;=======================================
;[ I. ASCII code <> 0 ]                 ;
          cmp       ah,TRIGCODE         ; trigger key?
          jz        Trig_wk             ;   yes, jump
                                        ;
          mov       dx,di               ; save di in dx in preparation for scan
          cmp       al,EXTKBCODE        ; extended keyboard code?
          jz        Ext_wk              ;   yes, jump

     ;===================================
     ;[ A. check for keys with duplicate ]
     ;[    ASCII value like Tab (vs. ^I) ]
                                        ;
          mov       di,offset DupCodes  ; search table of codes
          mov       cx,DUPNUMB
          repne
          scasw
          mov       di,dx               ; restore di
          jnz       ASCII_wk            ; not found, jump to write ASCII symbol
                                        ; found, cx = index into table
          mov       si,offset DupSymb   ; si = ptr to table of symbol addresses
Symb_wk
          shl       cx                  ; index --> offset into table
          add       si,cx               ; si = ptr to element of table
          mov       si,[si]             ; si = element = address of symbol
          jmps      Wri_wk              ; jump to write symbol

     ;=============================
     ;[ B. extended keyboard code ]
Ext_wk
          mov       di,offset ExtCodes  ; search table of codes
          mov       cx,EXTNUMB
          repne
          scasw
          mov       di,dx               ; restore di
          mov       si,offset ExtSymb   ; si = ptr to table of symbol addresses
          jz        Symb_wk             ; found. cx = index into table. jump
                                        ; not found, write '<E nnn>'
XNum_wk   ;------------------
          call      WriteXNum
          jmps      End_wk

     ;==================
     ;[ C. trigger key ]
Trig_wk
          mov       byte [di],'*'       ; write trigger-key symbol to TEXTBUFF
          inc       di                  ; advance buffer ptr
                                        ; write ASCII symbol
ASCII_wk  ;------------------
          or        Flag,KEY
          call      WriteASCI
          and       Flag,not KEY
          jmps      End_wk

;=======================================
; [ II. ASCII code = 0 ]
Spec_wk
          xchg      al,ah               ; al = scan code,  ah = 0
          cmp       al,MAXSCAN          ; scan code <= max number in table?
          ja        XNum_wk             ;   no, jump to write '<snnn>'

          call      CheckSpec
          jz        XNum_wk             ; if scan code not found, jump
Wri_wk
          movsb                         ; write symbol
          cmp       byte [si],0
          jnz       Wri_wk
End_wk
          mov       ax,'  '             ; write 2 spaces after key symbol
          stosw
          mov       ax,'= '             ; write equal sign + 1 space
          stosw                         ; 2nd space is written in WriteDef Proc.
          ret
          ENDP
;͸
;              WriteDef                
; writes definition symbols separated  
; by spaces to TEXTBUFF                
;                                      
;  In: bp = ptr to start of definition 
;      di = ptr to TEXTBUFF            
; Out: di updated                      
;      bp preserved                    
;;
WriteDef  PROC      NEAR
          mov       bx,bp               ; bx = ptr to definition in MACBUFF
          mov       cx,BytPerMac        ; cx = bytes remaining in definition

;[ loop through definition ]
Next_wd
          mov       al,[bx]             ; al = code
          or        al,al               ; end of definition?
          jz        End_wd              ;    yes, jump
          inc       bx                  ; advance ptr to definition in MACBUFF
          mov       byte [di],' '       ; write space to TEXTBUFF
          inc       di                  ; advance pointer to TEXTBUFF

          cmp       al,0FFh             ; special key?
          jz        Spec_wd             ;    yes, jump
          cmp       al,'#'              ; possibly #C, #c, #N, #n, #W, #w?
          jz        N#_wd               ;    yes, jump

     ;[ I. ASCII code ]
          call      WriteASCI
          loop      Next_wd
          jmps      End_wd

     ;[ II. special key: FFh ]
Spec_wd
          call      WriteFF
          loop      Next_wd
          jmps      End_wd

     ;[ III. # ]
N#_wd
          call      Write#
          loop      Next_wd

End_wd
          mov       ax,0A0Dh            ; write CR LF
          stosw
          ret
          ENDP
;͸
;             WriteASCI                
; writes symbol for ASCII values 1-254 
;                                      
;  In: al = ASCII value                
;      di = destination in TEXTBUFF    
; Out: di updated                      
;;
WriteASCI PROC      NEAR
          cmp       al,32               ;
          jb        Lo_wa               ; if al < 32, jump
          jz        Sp_wa               ; if al = 32, jump
          cmp       al,127              ;
          jz        B_wa                ; if al = 127, jump

     ;[ ASCII 33 - 126, 128 - 254 ]
          test      Flag,KEY            ; if key, don't double ' or "
          jnz       Sto2_wa

          cmp       al,27h              ; 27h = '   If ' or ", write '' or ""
          ja        Sto2_wa
          jz        Sto1_wa
          cmp       al,'"'
          jnz       Sto2_wa
Sto1_wa
          stosb                         ; store first ' or "
Sto2_wa
          stosb                         ; store character
          ret

     ;[ ASCII 0 - 31 ]
Lo_wa
          mov       ah,al               ; change  1 --> ^A
          add       ah,40h              ;         2 --> ^B  etc.
          mov       al,'^'
          stosw
          ret

     ;[ ASCII 127: ^BkSp ]
B_wa
          mov       si,offset CtlBkSp
          jmps      Wri_wa

     ;[ ASCII 32: Space ]
Sp_wa
          mov       si,offset Space
Wri_wa
          movsb                         ; write symbol
          cmp       byte [si],0
          jnz       Wri_wa
          ret
          ENDP
;͸
;               WriteFF                
;  In: bx = ptr to byte after FFh      
;      di = destination in TEXTBUFF    
; Out: di,bx,cx updated                
;;
WriteFF   PROC      NEAR
          mov       al,[bx]             ; al = scan code or special code
          inc       bx                  ; advance ptr to definition in MACBUFF
          dec       cx                  ; update bytes remaining
          xor       ah,ah               ; zero out ah

          cmp       al,MAXSCAN          ; al > number of elements in table?
          ja        High_wf             ;    yes, jump to check special codes

;=======================================
;[ I. use table of symbol addresses ]
          call      CheckSpec
          jz        XNum_wf

Symb_wf   ;-----------------
          movsb                         ; write symbol
          cmp       byte [si],0
          jnz       Symb_wf
          ret

;=======================================
;[ II. check for special codes ]
High_wf
          cmp       al,0E6h             ; special code for <Enter>?
          jz        Ent_wf              ;    yes, jump
          cmp       al,0E7h             ; special code for <BkSp>?
          jz        Bks_wf              ;    yes, jump;
          cmp       al,0E8h             ; special code for <Tab>?
          jz        Tab_wf              ;    yes, jump
          cmp       al,0E9h             ; special code for <Esc>?
          jnz       XNum_wf             ;    no, jump to write '<snnn>'

          mov       si,offset Esc
          jmps      Symb_wf
Ent_wf
          mov       si,offset Enter
          jmps      Symb_wf
Bks_wf
          mov       si,offset BkSp
          jmps      Symb_wf
Tab_wf
          mov       si,offset Tab
          jmps      Symb_wf

XNum_wf   ;------------------
          call      WriteXNum
          ret
          ENDP
;͸
;              CheckSpec               
; Checks for unrecognized scan code if 
; NewKey keys not recognized. Finds    
; address of symbol name in SpecSymb.  
;                                      
;  In: al = scan code                  
; Out: if symbol found                 
;         si = ptr to symbol name      
;         ZF clear                     
;      else ZF set                     
;;
CheckSpec PROC      NEAR
          mov       si,offset SpecSymb  ; si = ptr to table of symbol addresses
          shl       ax                  ; index --> offset into table
          add       si,ax               ; si = ptr to element of table (ZF clear)
          mov       si,[si]             ; si = element = address of symbol
          test      Flag,NEWKEY         ; use NewKey keys?
          jnz       Ret_cksn            ;   yes, ZF clear. jump

          shr       ax                  ; restore ax to original value
          mov       dx,di               ; save di in dx
          mov       di,si               ; di = start of symbol
          xchg      al,ah               ; al = 0. scan for null termination
          mov       cx,8                ; max symbol length + 1 for terminal 0
          repne
          scasb
          xchg      ah,al               ; restore al and ah
          cmp       byte [di],0FFh      ; byte after null termination = 0FFh?
          mov       di,dx               ;   if yes, NewKey key found, ZF set
Ret_cksn
          ret
          ENDP
;͸
;              Write#                  
;if # followed by X, for X=C,c,N,n,W,w 
;  writes '#X' to TEXTBUFF and updates 
;  bx and cx                           
; else writes '#'                      
;                                      
;   In: al = '#'                       
;       bx = ptr to next byte in macro 
;       di = ptr to TEXTBUFF           
;  Out: di (and maybe bx & cx) updated 
;;
Write#    PROC      NEAR
          mov       ah,[bx]             ; ah = code following '#'
          cmp       ah,'C'
          jz        A_w#
          cmp       ah,'N'
          jz        A_w#
          cmp       ah,'W'
          jz        A_w#
          cmp       ah,'c'
          jz        A_w#
          cmp       ah,'n'
          jz        A_w#
          cmp       ah,'w'
          jz        A_w#
          stosb                         ; write '#'
          ret
A_w#
          stosw                         ; write '#X' as one symbol
          inc       bx                  ; advance ptr to definition in MACBUFF
          dec       cx                  ; update bytes remaining
          ret
          ENDP
;͸
;             WriteXNum                
;writes <ennn> or <snnn> to TEXTBUFF   
;                                      
;In: Scan:     ah:al = 0:number        
;    Extended: ah:al = number:EXTKBCODE
;    di = ptr to TEXTBUFF              
;Out: di updated                       
;;
WriteXNum PROC      NEAR
          mov       dx,ax               ; save number in dx
          mov       ax,'<s'             ; initialize ax to '<s' for scan code
          or        dh,dh               ; extended code?
          jz        Go_wxn              ;   no, jump
          mov       ah,'e'              ;   yes, change 's' to 'e'
          mov       dl,dh               ;        put number to write in dl
Go_wxn
          stosw                         ; write '<s' or '<e'
          mov       al,dl               ; al = number to write
          call      WriteNum            ; write number
          mov       al,'>'              ; write '>'
          stosb
          ret
          ENDP
;͸
;              WriteNum                
;   writes number to TEXTBUFF, where   
;   0 <= number <= 255                 
;                                      
;  In: al = unsigned number            
;      di = ptr to TEXTBUFF            
; Out: di updated                      
;;
WriteNum  PROC      NEAR
          mov       dl,10               ; dl = divisor
          xor       dh,dh               ; dh = number of digits

     ;[ store digits on stack ]
Get_wn
          xor       ah,ah               ; zero out remainder
          div       dl                  ; al = al/10 with remainder in ah
          push      ax                  ; save remainder digit on stack
          inc       dh                  ; count one digit
          or        al,al               ; any more?
          jnz       Get_wn              ;   yes, loop back
                                        ;   no, write digits
     ;[ write digits to buffer ]
Put_wn
          pop       ax                  ; pop digit off stack
          mov       al,ah               ; move digit to al
          add       al,'0'              ; change to ASCII code
          stosb                         ; write to destination
          dec       dh                  ; decrement count
          jnz       Put_wn              ; jump if more digits

          ret
          ENDP
; TEXT-TO-MACRO 
;͸
;             TextToMac                
;       converts text to macros        
;                                      
;  In: TEXTBUFF contains text file     
;      cx = bytes read into TEXTBUFF   
; Out: di = limit of data in MACBUFF   
; Err: NOWRITE flag set                
;;
TextToMac PROC      NEAR
          mov       pLineMark,TEXTBUFF  ; initialize to start of TEXTBUFF
          mov       RemBytes,cx         ; store bytes read into TEXTBUFF
          call      UpperCase           ; change all symbols to upper case

;=======================================
;[ I. write number of bytes/macro ]
          call      GetLine             ; si = ptr to bytes/macro
          jc        Xu_ttm              ;   if no lines, jump to show usage
          call      StrToNum            ; convert to number in ax
          jc        Xb_ttm              ;    jump if problem
          cmp       ax,5                ; ax < 5?
          jb        Xb_ttm              ;   yes, error. jump
          cmp       ax,96               ; ax > 96?
          ja        Xb_ttm              ;   yes, error. jump
          mov       BytPerMac,ax        ; store
          mov       di,MACBUFF          ; di = destination in MACBUFF
          stosw                         ; write to MACBUFF
          mov       pMacMark,di         ; store MACBUFF position

;=======================================
;[ II. convert text lines to macros ]
          call      GetLine             ; si points to next data line = 1st mac
          jc        Xu_ttm              ;   if no lines, jump to show usage

Next_ttm
          call      ConvLine            ; convert text line to macro
          cmp       di,MBUFFEND         ; beyond end of MACBUFF?
          ja        Xm_ttm              ;   yes, too many macros. jump
          call      GetLine             ; any more data lines?
          jnc       Next_ttm            ;   yes, jump to convert

;=======================================
;[ III. add terminal 0 ]
          mov       di,pMacMark         ; di = position in MACBUFF
          xor       ax,ax
          stosw                         ; write 0
          ret                           ; out: di = limit of macros in MACBUFF
;---------------------------------------
     ;[ error handling ]
Xb_ttm
          mov       si,offset Xbmsg     ; "bad bytes/macro" message
          jmps      X_ttm
Xm_ttm
          mov       si,offset Xmmsg     ; "too many macros" message
          jmps      X_ttm
Xu_ttm
          mov       si,offset Umsg      ; usage message
X_ttm
          call      ErrorExit
          ENDP
;͸
;              UpperCase               
;    Changes all symbols in tables     
;    to upper case                     
;;
UpperCase PROC      NEAR
     ;[ 2-character symbols]
          mov       si,offset s67 + 5
          mov       cx,15
A_uc
          mov       byte [si],'N'
          add       si,5
          loop      A_uc

          mov       byte [offset Space + 1],'P'
          and       byte [offset s72 + 1],'P'

     ;[ 3-character symbols ]
          mov       byte [offset s168 + 1],'N'
          mov       byte [offset s167 + 1],'N'
          mov       byte [offset s154 + 2],'P'

          mov       bx,offset XTable

          mov       si,offset s112 + 6
          mov       dx,3
          mov       bp,16
          call      UpperWrd

          mov       byte [offset s150 + 1], 'N'
          mov       byte [offset s144 + 1], 'N'
          mov       byte [offset s142 + 1], 'N'
          mov       byte [offset s149 + 1], 'N'
          mov       byte [offset s12 + 2],'P'
          mov       byte [offset s141 + 2],'P'

     ;[ 4-character symbols]
          mov       si,offset FourTabl
          inc       dx                  ; dx = 4
          mov       bp,FOURNUMB
          call      UpperWrd

     ;[ 5-character symbols]
          mov       si,offset FiveTabl
          inc       dx                  ; dx = 5
          mov       bp,FIVENUMB
          call      UpperWrd

     ;[ 6-character symbols]
          mov       si,offset SixTabl
          inc       dx                  ; dx = 6
          mov       bp,SIXNUMB
          call      UpperWrd

     ;[ 7-character symbols]
          mov       si,offset SevenTabl
          inc       dx                  ; dx = 7
          mov       bp,SEVENNUMB
          call      UpperWrd

          ret
          ENDP
;͸
;               UpperWrd               
;         Helper for UpperCase         
;                                      
; In:  si = ptr into symbol table      
;      dx = length of symbol           
;      bp = number of symbols to change
;      bx = ptr to Xtable              
;;
UpperWrd  PROC      NEAR
          mov       di,si
          mov       cx,dx
A_upw
          lodsb
          xlat
          stosb
          loop      A_upw
          add       si,3                ; skip terminal zero and codes
          dec       bp                  ; decrement count
          jnz       UpperWrd            ; if not done, loop back
          ret
          ENDP
;͸
;              GetLine                 
;gets start of next data line, skipping
;   comment lines and blank lines      
;puts 0 at limit of line (i.e. changes 
;   line to ASCIZ string).             
;                                      
;  In: pLineMark = ptr to TEXTBUFF     
;      RemBytes = bytes remaining      
; Out: si = pLine = ptr to next line   
;      pLineMark = limit of line       
;      RemBytes updated                
;      CF clear                        
;Else: CF set if end of buffer         
;;
GetLine   PROC      NEAR
          mov       di,pLineMark        ; starting place in TEXTBUFF
          mov       cx,RemBytes         ; bytes remaining in TEXTBUFF
          mov       al,LF

;=======================================
;[ I. find start of next line ]

     ;[ scan to next LF ]
Nxt_gtl
          repne
          scasb

     ;[ check for valid data line ]
          jnz       None_gtl            ; if LF not found, no next line. done
          cmp       cx,5                ; less than 5 bytes left (i.e.
                                        ;   char + space + '=' + space + def)?
          jb        None_gtl            ;   yes, can't be any more macros. done
          mov       dx,[di]             ; dl = [di], dh = [di+1]
          cmp       dl,CR               ; blank line?
          jz        Nxt_gtl             ;    yes, jump to try again
          cmp       dl,' '              ; comment line?
          jz        Nxt_gtl             ;    yes, jump to try again
          cmp       dl,';'              ; first character = ';'?
          jnz       Start_gtl           ;   no, found start, jump
          cmp       dh,' '              ;   yes. second character = space?
          jnz       Nxt_gtl             ;        no, it's a comment, try again
                                        ;        yes, found start
     ;[ save start ]
Start_gtl
          mov       si,di               ; out: si = ptr to line
          mov       pLine,si            ; store start for writing error message

;=======================================
;[ II. find limit of line ]
          repne                         ; scan to next LF
          scasb
          jnz        NoLF_gtl           ; if LF not found, jump

          sub       di,2                ; go back to CR
          add       cx,2                ; adjust remaining bytes

Set_gtl
          mov       byte [di],0         ; change limit of line to 0
          mov       pLineMark,di        ; store position for next call
          mov       RemBytes,cx         ; store remaining bytes
          clc
          ret

     ;[ LF not found: end of buffer ]
NoLF_gtl
          cmp       byte [di-1],EOF     ; is EOF last character in buffer?
          jnz       Set_gtl             ;    no, jump
          dec       di                  ;    yes, move back one byte to EOF
          jmps      Set_gtl

;---------------------------------------
None_gtl
          stc
          ret
          ENDP
;͸
;              ConvLine                
;     converts text line to macro      
;                                      
;  In: si = start of 0-terminated      
;           data line                  
;      pMacMark = position in MACBUFF  
; Out: pMacMark = di = new position in 
;                      MACBUFF         
;;
ConvLine  PROC      NEAR
          mov       ax,pMacMark         ; position in MACBUFF = start of macro
          mov       pMacStart,ax        ; store start for length calculation
          mov       pSymMark,si         ; initialize pSymMark to start of line

;=======================================
;[ I. key ]

     ;[ get key codes ]
          or        Flag,KEY            ; set flag
          call      GetSymbol           ; get key symbol in SYMBUFF
          jc        Xk_cvln             ; if none or bad symbol, error. jump
          call      GetCodes            ; symbol --> scan & ASCII codes in ax
          jc        Len_cvln            ; jump on error (don't need to clear flag)
          and       Flag,not KEY        ; clear flag

          mov       di,pMacMark         ; position in MACBUFF = start of macro

     ;[ check if key already used ]
          mov       si,di               ; si = position of key in MACBUFF
          mov       dx,BytPerMac        ; dx = bytes allotted to one macro
Chk_cvln
          sub       si,dx               ; si = ptr to previous key, if any
          cmp       si,MACBUFF          ; reached start of buffer?
          jbe       Sto_cvln            ;   yes, jump out of loop
          cmp       [si],ax             ; already used?
          jnz       Chk_cvln            ;   no, jump to check previous key
          jmps      Xa_cvln             ;   yes, error. jump

     ;[ write key codes to buffer ]
Sto_cvln
          stosw
          mov       pMacMark,di         ; update buffer position

;=======================================
;[ II. equal sign ]
          call      GetSymbol           ; get next symbol. ax = length.
          jc        Xe_cvln             ; if none or bad symbol, error. jump
          cmp       byte [SYMBUFF],'='  ; is symbol an equal sign?
          jnz       Xee_cvln            ;   no, error. jump
          dec       ax                  ; more than one character?
          jnz       Xee_cvln            ;   yes, error. jump

;=======================================
;[ III. definition loop ]
          call      GetSymbol           ; get first symbol of definition
          jc        Xd_cvln             ; if none or bad symbol, error. jump
Next_cvln
          test      Flag,STRING
          jnz       Cod_cvln
          call      GetCodes            ; symbol --> scan & ASCII codes in ax
          jc        Len_cvln            ; jump on error
Cod_cvln
          call      DefCodes            ; get definition codes and write to buff
          call      GetSymbol           ; any more symbols?
          jnc       Next_cvln           ;   yes, loop back
                                        ;   if no symbols, end of def. OK.
                                        ;   if bad symbol, message already written
;=======================================
;[ IV. check length of macro ]
Len_cvln
          mov       di,pMacMark         ; current position in MACBUFF
          mov       ax,di               ; ax = limit of macro
          sub       ax,pMacStart        ; ax = limit - start = length
          mov       cx,BytPerMac        ; cx = bytes allotted per macro
          sub       cx,ax               ; cx = unused bytes in allotted space
          jle       Xt_cvln             ; if 0 or negative, macro is too long

;=======================================
;[ V. fill rest of macro with 0's ]
Zero_cvln
          xor       al,al
          rep
          stosb
          mov       pMacMark,di         ; update position in MACBUFF
          ret                           ; out: di = position in MACBUFF
;---------------------------------------
     ;[ error handling ]
Xk_cvln
          jnz       Len_cvln            ; if bad symbol, message written. jump
          mov       si,offset Xkmsg     ; "bad key" message
          jmps      X_cvln              ; jump to write

Xa_cvln
          mov       si,offset Xamsg     ; "key already used" message
          jmps      X_cvln              ; jump to write

Xd_cvln
          jnz       Len_cvln            ; if "bad symbol" message written. jump
          mov       si,offset Xdmsg     ; "no definition" message
          jmps      X_cvln              ; jump to write

Xe_cvln
          jnz       Len_cvln            ; if "bad symbol" message written. jump
Xee_cvln
          mov       si,offset Xemsg     ; "problem with equal sign" message

X_cvln
          call      LineError           ; write line and message
          jmps      Len_cvln            ; continue processing file

Xt_cvln
          mov       si,offset Xtmsg     ; "macro too long" message
          call      LineError           ; write line and message
          mov       di,pMacStart        ; di = start of macro
          add       di,2                ; di = start of definition
          mov       cx,BytPerMac        ; number of zeros to write
          jmps      Zero_cvln           ; jump to zero out definition
          ENDP
;͸
;             GetSymbol                
; copies symbol from current line of   
; TEXTBUFF to SYMBUFF                  
;                                      
;  In: pSymMark = starting position in 
;                 line                 
; Out: SYMBUFF contains symbol         
;      ax = length of symbol           
;      pSymMark = limit of symbol      
;      STRING flag set if string       
;      CF clear                        
;Else: CF set & ZF set if no symbols   
;      i.e. if comment, or end of line 
; Err: CF set & ZF clear               
;;
GetSymbol PROC      NEAR
          mov       si,pSymMark

;=======================================
;[ I. find start ]
A_gtsy
     ;[ skip spaces ]
          lodsb                         ; al = character
          or        al,al               ; end of line?
          jz        NoMore_gtsy         ;   yes, no symbol. jump to set CF
          cmp       al,' '              ; al = space?
          je        A_gtsy              ;   yes, skip

     ;[ check for semicolon ]
          cmp       al,';'
          jz        Com_gtsy            ; if ';', jump to check if comment

     ;[ check for quotation mark ]
          cmp       al,'"'
          jz        Qmrk_gtsy
          cmp       al,27h              ; 27h = '
          jnz       SavS_gtsy           ; if ' or " not found, jump to save
                                        ;    start
     ;[ ' or " found ]
Qmrk_gtsy
          test      Flag,KEY            ; if key, treat as single character
          jnz       SavS_gtsy           ;   so, jump to save start

          cmp       al,[si]             ; "" or '' ?
          jz        Dbl_gtsy            ;   yes, jump
          call      GetString           ;   no, set up to copy string to SYMBUFF
          jnc       Cpy_gtsy
          jmps      XX_gtsy

     ;[ "" or '' ]
Dbl_gtsy
          mov       cx,si               ; cx = start of symbol
          inc       si
          jmps      Lim_gtsy            ; jump to find limit

     ;[ check for comment ]
Com_gtsy
          cmp       byte [si],' '       ; next character = space?
          jz        SavS_gtsy           ;   yes, not comment, jump
          cmp       byte [si],0         ; is ';' last character in line?
          jnz       NoMore_gtsy         ;   no, comment found, jump to set CF

     ;[ save start ]
SavS_gtsy
          mov       cx,si               ; si = one byte past start of symbol
          dec       cx                  ; cx = start of symbol

;=======================================
;[ II. find limit ]
          cmp       al,'<'              ; <xxxxx> type of symbol?
          jnz       Lim_gtsy            ;   no, jump to find limit
          cmp       byte [si],' '       ; individual character?
          jz        Lim_gtsy            ;   yes, jump to find limit
          cmp       byte [si+1],' '     ; <s nnn> or <e nnn>?
          jnz       Lim_gtsy            ;   no, jump to find limit
          add       si,2                ;   yes, advance to byte beyond space

Lim_gtsy
          lodsb                         ; al = character
          or        al,al               ; end of line?
          jz        SavL_gtsy           ;   yes, found limit. jump to save
          cmp       al,' '              ; al = space?
          jnz       Lim_gtsy            ;   yes, jump to advance

     ;[ store limit ]
SavL_gtsy
          dec       si                  ; go back to limit of symbol
          mov       pSymMark,si         ; store limit for next call

;=======================================
;[ III. check length ]
          xchg      si,cx               ; si = start of symbol, cx = limit
          sub       cx,si               ; cx = length
          cmp       cx,7                ; longest possible symbol is 7 bytes
          ja        X_gtsy              ; if cx > 7, error. jump
                                        ;    else copy to SYMBUFF
;=======================================
;[ IV. copy to SYMBUFF ]
Cpy_gtsy
          mov       ax,cx               ; out: ax = length
          mov       di,SYMBUFF
          rep
          movsb
          mov       [di],cl             ; make ASCIZ string (cx = 0 after rep)
          clc                           ; CF clear = symbol found
          ret

;=======================================
;[V. no symbol found ]
NoMore_gtsy
          xor       ax,ax               ; ZF set as end-of-definition signal,
          stc                           ;    i.e. not bad symbol
          ret                           ; CF set = no symbol available

;---------------------------------------
     ;[ error handling ]
X_gtsy
          mov       bp,cx               ; save length of symbol in bp
          call      PutLine             ; write text line to STDOUT

          mov       dx,offset Indent    ; write 4 spaces + "
          mov       cx,INDENTLEN
          call      WriteOut

          mov       dx,si               ; dx = ptr to symbol
          mov       cx,bp               ; cx = length of symbol
          call      WriteOut

          mov       si,offset Xsmsg     ; "bad symbol" message
          call      PutMsg
XX_gtsy
          or        Flag,NOWRITE        ; ZF clear as bad-symbol signal
          stc                           ; CF set = no symbol available
          ret
          ENDP
;͸
;              GetString               
;                                      
; In: al = ' or "                      
;     si = ptr to char. after ' or "   
;Out: cx = length of string            
;     si preserved                     
;     STRING flag set                  
;     CF clear                         
;Err: CF set                           
;;
GetString PROC      NEAR
          mov       di,si               ; si = 1st character of string
          mov       cx,pLineMark        ; limit of line
          sub       cx,di               ; cx = bytes remaining in line
Nxt_gtst
          repne                         ; look for matching quotation mark
          scasb
          jnz       Xq_gtst             ; if not found, error. jump
          cmp       byte [di],' '       ; is next character a space?
          jz        Set_gtst            ;   yes, OK. jump.
          cmp       byte [di],0         ; end of the line?
          jnz       Nxt_gtst            ;   no, not end of string. try again.
Set_gtst
          mov       pSymMark,di         ; set mark for next call to GetSymbol
          mov       cx,di               ; limit of string + quotation mark
          sub       cx,si               ; length of string + quotation mark
          dec       cx                  ; out: cx = length of string
          jz        Xq_gtst             ; if length = 0, error. jump
          cmp       cx,BytPerMac        ; rough check of length
          jae       Xt_gtst
          or        Flag,STRING         ; set flag, CF clear
          ret
;---------------------------------------
     ;[ error handling ]
Xq_gtst
          mov       si,offset Xqmsg
          jmps      X_gtst
Xt_gtst
          mov       si,offset Xtmsg
X_gtst
          call      PutLine
          call      PutMsg
          stc
          ret
          ENDP
;͸
;              GetCodes                
; returns scan & ASCII codes for symbol
; in SYMBUFF                           
;                                      
;  In: ax = length of symbol           
;      KEY flag set if symbol is macro 
;          key                         
; Out: ah = scan code (or trigger code)
;      al = ASCII code                 
;      CF clear                        
; Err: CF set                          
; ------------------------------------ 
; Possibilities for first character    
; of multi-character symbol:           
;   # (35)  = #C, #N, #W, #c, #n       
;   $ (36)  = Shift                    
;   * (42)  = Trigger Key              
;   < (60)  = numerical value          
;   @ (64)  = Alt                      
;   Upper case                         
;   ^ (94)  = Ctrl                     
;   Lower case                         
;;
GetCodes  PROC      NEAR
          cmp       ax,1                ; only one character?
          jnz       Mul_gtcd            ;   no, jump

;=======================================
;[ I. one character ]
          mov       al,[SYMBUFF]        ; al = ASCII code
          xor       ah,ah               ; set scan code to 0 initially
          test      Flag,KEY            ; is code for macro key?
          jnz       K_gtcd              ;   yes, jump
          ret                           ;   no, done
K_gtcd
          call      GetScan             ; for key, get scan code in ah
          jc        X_gtcd
          ret

;=======================================
;[ II. multi-character symbol ]
Mul_gtcd
          mov       dx,[SYMBUFF]        ; dl = first char, dh = second

     ;[<<DECISIONS>>]
          cmp       dl,'^'              ; <Ctrl> ?
          jz        Ctrl_gtcd           ;   yes, jump
          cmp       dl,'@'
          ja        Bs_gtcd             ; if dl > '@', jump to binary search
          jb        Lo_gtcd             ; if dl < '@', jump to clarify
          cmp       ax,2                ; dl = '@'.  more than 2 characters?
          ja        Bs_gtcd             ;   yes, jump to binary search

     ;=======================================
     ;[ A. <Alt x>, where x = any character ]
          mov       al,dh               ;  al = character
          mov       bx,offset ScanTable ;  get scan code from table
          xlat
          cmp       al,13               ; scan code > 13 ?
          ja        A2_gtcd             ;   yes, OK. jump
          add       al,076h             ; for top row of keyboard (<1> - <=>)
A2_gtcd                                 ;    add 76h to get <Alt x> scan code
          mov       ah,al               ; ah = scan code
          xor       al,al               ; al = ASCII code = 0  (CF clear)
          ret

     ;[<<DECISIONS>>]
Ctrl_gtcd                               ; 1st character = '^'
          cmp       ax,2                ;  more than 2 characters
          jnz       Bs_gtcd             ;   yes, jump to binary search
          cmp       dh,'@'              ; 2nd character < '@' ?
          jb        Bs_gtcd             ;   yes, jump to binary search
          cmp       dh,125              ; 2nd character > 125?
          ja        Bs_gtcd             ;   yes, jump to binary search
          cmp       dh,'`'              ; 2nd character = '`'?
          jz        Bs_gtcd             ;   yes, jump to binary search
          and       dh,05Fh             ; capitalize

     ;======================================
     ;[ B. <Ctrl x>, where 'A' <= x <= '_' ]
          mov       al,dh               ; al = character
          mov       bx,offset ScanTable ; get scan code from table
          xlat
          mov       ah,al               ; ah = scan code
          mov       al,dh               ; al = character
          sub       al,040h             ; ASCII code = char - 040h (CF clear)
          ret                           ;   i.e   ^A --> 1
                                        ;         ^B --> 2  etc.

;---------------------------------------
     ;[ error handling ]                ;
X_gtcd                                  ;
          call      SymError            ;
          ret                           ;
;---------------------------------------

     ;====================================
     ;[ C. binary search of symbol table ]
Bs_gtcd
          mov       si,ax               ; 2 <= si <= 7  (length of symbol)
          sub       si,2                ; 0 <= si <= 5  for 0-based indexing
          shl       si                  ; index --> ...
          shl       si                  ;    ...offset into list ( 4 bytes/elem)
          add       si,offset SymTabAdd ;  + list of addresses of symbol tables
          mov       bx,[si]             ; bx = ptr to first element of table
          mov       bp,[si+2]           ; bp = ptr to last element of table
          call      BinSearch
          jc        X_gtcd
          ret

     ;[<<DECISIONS>>]
Lo_gtcd
          cmp       dl,'$'
          jz        Bs_gtcd             ; if dl = '$', jump to binary search
          jb        N#_gtcd
          cmp       dl,'*'
          ja        Num_gtcd

     ;==================
     ;[ D. Trigger Key ]
          call      ConvTrig
          jc        X_gtcd
          ret

     ;==============
     ;[ E. <xxxxx> ]
Num_gtcd
          call      ConvNumb
          jc        X_gtcd
          ret

     ;=========
     ;[ F. #? ]
N#_gtcd
          call      Conv#
          jc        X_gtcd
          ret
          ENDP
;͸
;              GetScan                 
;    gets scan code for macro key      
;                                      
;  In: al = ASCII value                
; Out: al preserved                    
;      ah = scan code                  
;      CF clear                        
; Err: CF set                          
;;
GetScan   PROC      NEAR
          or        al,al
          js        X_gtsc              ; if al > 127, error. jump

          mov       bx,offset ScanTable
          mov       ah,al               ; save ASCII value in ah
          xlat                          ; al = scan code
          xchg      ah,al               ; out: ah = scan, al = ASCII
          clc
          ret
;---------------------------------------
     ;[ error handling ]
X_gtsc
          stc
          ret
          ENDP
;͸
;             BinSearch                
;  binary search procedure for symbol  
;  tables                              
;                                      
; Note: for speed, this procedure      
; does not use a StrCmp subroutine,    
; which would have to be called many   
; times                                
;                                      
; In: bx = ptr to first elem of table  
;     bp = ptr to last elem of table   
;     SYMBUFF contains target string   
;     ax = length of target            
;                                      
;Out: ah:al = scan:ASCII codes         
;     CF clear                         
;Err: CF set                           
;;
BinSearch PROC      NEAR
          mov       SymbLen,ax          ; bytes to check
          mov       ElemLen,ax
          add       ElemLen,3           ; ElemLen = length of table element
                                        ;    = stringlen + '0' + 2 bytes for codes
     ;[ copy and capitalize ]
          mov       si,SYMBUFF          ; si = ptr to symbol
          mov       di,BSBUFF           ; di = ptr to destination buffer
          mov       cx,ax               ; cx = length of symbol
          mov       dx,bx               ; save bx in dx
          mov       bx,offset XTable    ; bx = ptr to table to change to upper
                                        ;      case using xlat
Copy_bs
          lodsb
          xlat
          stosb
          loop      Copy_bs
          mov       bx,dx               ; restore bx from dx

     ;[ check first element of symbol table ]
          mov       si,bx
          mov       di,BSBUFF           ; di = ptr to target
          mov       cx,SymbLen          ; cx = bytes to check
          repe
          cmpsb
          jb        Start_bs            ; if left < target, jump into loop
          jz        OK_bs               ; if match, done
          jmps      X_bs                ; if target < left, error

     ;[ search loop ]
Left_bs
          mov       bx,ax               ; middle < target, so new left = middle
Start_bs
          mov       ax,bp               ; ax = bp = right
Next_bs
          sub       ax,bx               ; ax = right - left
          cmp       ax,ElemLen          ; ax <= ElemLen ?
          jbe       Last_bs             ;    yes, jump to last check
          shr       ax                  ; ax = (right-left)/2
          mov       si,ax               ; save in si
          xor       dx,dx
          div       ElemLen             ; dx = rem( si / ElemLen )
          sub       si,dx               ; si = multiple of ElemLen
          add       si,bx               ; si = middle
          mov       ax,si               ; save middle in ax

          mov       di,BSBUFF           ; di = ptr to target
          mov       cx,SymbLen          ; cx = bytes to check
          repe
          cmpsb
          jb        Left_bs             ; if middle < target, jump to set left
          jz        OK_bs               ; if match, done
          mov       bp,ax               ; target < middle, so new right = middle
          jmps      Next_bs

     ;[ final check on right ]
Last_bs
          mov       si,bp               ; si = right
          mov       di,BSBUFF           ; di = ptr to target
          mov       cx,SymbLen          ; cx = bytes to check
          repe
          cmpsb
          jnz       X_bs
OK_bs                                   ; CF clear because cmpsb found match
          inc       si                  ; skip past null-termination of string
          mov       ax,[si]             ; ax = codes
          inc       al                  ; al = 0FFh (i.e. NewKey key) ?
          jz        NK_bs               ;   yes, jump to check Flag
          dec       al                  ; restore al to previous value
          ret                           ; CF clear because mov, inc, and dec
                                        ;    don't affect CF
NK_bs
          test      Flag,NEWKEY         ; use NewKey keys?  (CF clear)
          jz        X_bs                ;   no, jump to set CF
          ret
;---------------------------------------
     ;[ error handling ]
X_bs
          stc
          ret
          ENDP
;͸
;              ConvTrig                
; converts '*X' symbol in SYMBUFF to   
; trigger and ASCII codes              
;                                      
; In: dl should contain '*'            
;     dh = next character              
;     ax = length                      
;     KEY flag set if symbol is macro  
;         key                          
;                                      
;Out: ah = TRIGCODE (0E6h)             
;     al = ASCII code                  
;     CF clear                         
;Err: CF set                           
;;
ConvTrig  PROC      NEAR
          cmp       dl,'*'              ; trigger key symbol?
          jnz       X_cvtg              ;   no, error. jump
          test      Flag,KEY            ; is this symbol for macro key?
          jz        X_cvtg              ;   no, error. jump
          cmp       ax,2                ; 2-byte symbol?
          jnz       X_cvtg              ;   no, error. jump
          cmp       dh,127              ; dh >= 127?
          jae       X_cvtg              ;   yes, error. jump
          cmp       dh,32               ; dh <= 32?
          jbe       X_cvtg              ;   yes, error. jump
          mov       al,dh               ; out: al = ASCII value
          mov       ah,TRIGCODE         ;      ah = trigger key code
          clc
          ret
;---------------------------------------
     ;[ error handling ]
X_cvtg
          stc
          ret
          ENDP
;͸
;              ConvNumb                
;converts <nnn> <e nnn> <ennn> <s nnn> 
;or <snnn> symbol in SYMBUFF to scan   
;and ASCII codes                       
;                                      
; In: dl should be '<'                 
;     dh = next character              
;     KEY flag set if symbol is macro  
;         key                          
; Out:                                 
;for <nnn>:  ah = scan code if key     
;                 0 if definition      
;            al = ASCII value          
;for <snnn> <s nnn>: ah = scan code    
;                    al = 0            
;for <ennn> <e nnn>: ah = scan code    
;                    al = EXTKBCODE    
;CF clear                              
;                                      
;Err: CF set                           
;;
ConvNumb  PROC      NEAR
          cmp       dl,'<'
          jnz       X_cvn
          cmp       dh,'9'              ; dh <= '9'?
          jbe       ASC_cvn             ;   yes, <nnn>. jump
          and       dh,05Fh             ; change dh to upper case
          cmp       dh,'E'
          jz        SE_cvn              ; if dh = 'E', jump to process
          cmp       dh,'S'              ; dh = 'S'?
          jnz       X_cvn               ;   no, error. jump

     ;[ <s nnn> or <e nnn> ]
SE_cvn
          mov       [SYMBUFF+1],dh      ; store upper case character in buffer
          mov       si,SYMBUFF+2        ; si = ptr to character after 'S' or 'E'
          cmp       byte [si],' '       ; space?
          jnz       S2N_cvn             ;   no, si points to number. jump
          inc       si                  ;   yes, advance si to start of number
S2N_cvn
          call      StrToNum             ; out: al = number, ah = 0
          jc        X_cvn                ; jump on error
          xchg      ah,al                ; ah = scan code, al = 0
          test      Flag,KEY             ; is this for a macro key?
          jz        Bra_cvn              ;   no, jump to check for '>'
          cmp       byte [SYMBUFF+1],'S' ; special?
          jz        Bra_cvn              ;   yes, jump to check for '>'
          mov       al,EXTKBCODE         ;   no, al = extended keyboard code
          jmps      Bra_cvn              ; jump to check for '>'

     ;[ <nnn> ]
ASC_cvn
          mov       si,SYMBUFF+1        ; si = ptr to start of number
          call      StrToNum            ; al = number, ah = 0
          jc        Ret_cvn
          test      Flag,KEY            ; is symbol for macro key?
          jz        Bra_cvn             ;   no, jump to check for '>'
          call      GetScan             ;   yes, get scan code in ah
          jc        Ret_cvn

     ;[ check for right bracket ]
Bra_cvn                                 ; after StrToNum, si points to 2 bytes
          cmp       byte [si-1],'>'     ;    after last digit in symbol
          jnz       X_cvn               ; if no right bracket, bad symbol
          ret                           ; CF clear if ZF set
;---------------------------------------
     ;[ error handling ]
X_cvn
          stc
Ret_cvn
          ret
          ENDP
;͸
;               Conv#                  
; for #C, #c, #N, #n, #W, #w in SYMBUFF
; writes '#' to MACBUFF and returns    
; ASCII code for next character        
;                                      
; In: dl should contain '#'            
;     dh = next character              
;     ax = length of symbol            
;     pMacMark = position in MACBUFF   
;     KEY flag set if symbol is macro  
;         key                          
;                                      
;Out: al = ASCII code of next character
;     ah = 0 since it cannot be for key
;     pMacMark updated                 
;     CF clear                         
;Err: CF set                           
;;
Conv#     PROC      NEAR
          cmp       dl,'#'              ; dl = '#'?
          jnz       X_cv#               ;   no, error. jump
          cmp       ax,2                ; length = 2?
          jnz       X_cv#               ;   no, error. jump
          test      Flag,KEY            ; symbol for key?
          jnz       X_cv#               ;   yes, error. jump

          cmp       dh,'C'              ; dh = 'C'?
          jz        CNW_cv#             ;   yes, jump to process
          cmp       dh,'N'              ; dh = 'N'?
          jz        CNW_cv#             ;   yes, jump to process
          cmp       dh,'W'              ; dh = 'W'?
          jz        CNW_cv#             ;   yes, jump to process
          cmp       dh,'c'              ; dh = 'c'?
          jz        CNW_cv#             ;   yes, jump to process
          cmp       dh,'n'              ; dh = 'n'?
          jz        CNW_cv#             ;   yes, jump to process
          cmp       dh,'w'              ; dh = 'w'?
          jnz       X_cv#               ;   no, error. jump
CNW_cv#
          mov       di,pMacMark         ; di = position in MACBUFF
          mov       [di],dl             ; write '#'
          inc       di                  ; advance position in MACBUFF
          mov       pMacMark,di         ; store position
          mov       al,dh               ; out: al = C, c, N, n, W, or w
          xor       ah,ah               ;      ah = 0, CF clear
          ret
;---------------------------------------
     ;[ error handling ]
X_cv#
          stc
          ret
          ENDP
;͸
;              StrToNum                
;  In: si = ptr to string of decimal   
;           digits.  max: 255          
; Out: al = 0 < number < 256           
;      ah = 0                          
;      si = 2 bytes past last digit    
;      CF clear                        
; Err: CF set                          
;;
StrToNum  PROC      NEAR
          xor       dx,dx               ; initialize number to 0
          mov       bl,10               ; base 10

Next_stn
          lodsb
          sub       al,'0'              ; ASCII --> decimal
          js        Done_stn            ; if less than '0', that's all.
          cmp       al,9                ; al <= 9?
          ja        Done_stn            ;    no, done

          cbw                           ; zero out ah
          xchg      ax,dx               ; dl = new digit, al = old number
          mul       bl                  ; ax = old number * 10
          jc        Y_stn               ; if ah <> 0, number is too big
          add       dl,al               ; dl = new digit + (old number * 10)
          jnc       Next_stn            ; if dl <= 255, jump to get next digit
X_stn                                   ; else error
          stc
Y_stn
          ret

Done_stn
          or        dx,dx               ; is number 0?   (CF clear)
          jz        X_stn               ;    yes, error. jump
          mov       ax,dx               ; out: ax = number
          ret
          ENDP
;͸
;              DefCodes                
; converts scan:ASCII codes to codes   
; for definition and writes to MACBUFF 
;                                      
; In -- no string:                     
;     ah = scan code                   
;     al = ASCII or extended kb code   
;     pMacMark = position in MACBUFF   
; In -- string:                        
;     ax = length of string            
;     SYMBUFF contains string          
;                                      
;Out: pMacMark updated                 
;;
DefCodes  PROC      NEAR
          mov       di,pMacMark         ; di = destination in MACBUFF
          test      Flag,STRING
          jnz       Str_dcd             ; if string, jump

          or        ah,ah               ; ah = scan code = 0?
          jz        Byt_dcd             ;   yes, jump to write ASCII code
          or        al,al               ; al = ASCII code = 0?
          jz        FF_dcd              ;   yes, jump to change al to FFh
          cmp       al,EXTKBCODE        ; al = extended keyboard code?
          jz        FF_dcd              ;   yes, jump to change al to FFh

     ;[ special handling required? ]
          cmp       ax,0E08h            ; <BkSp>?
          jz        Bks_dcd             ;   yes, jump
          cmp       ax,0F09h            ; <Tab>?
          jz        Tab_dcd             ;   yes, jump
          cmp       ax,011Bh            ; <Esc>?
          jz        Esc_dcd             ;   yes, jump
          cmp       ax,1C0Dh            ; <Enter>?
          jz        Ent_dcd             ;   yes, jump
                                        ;   no, just save ASCII value
Byt_dcd   ;--------------------
          stosb                         ; write ASCII code to MACBUFF
          mov       pMacMark,di         ; store position in MACBUFF
          ret

     ;[ convert to special codes ]
Bks_dcd
          mov       ah,0E7h             ; <BkSp>
          jmps      FF_dcd
Ent_dcd
          mov       ah,0E6h             ; <Enter>
          jmps      FF_dcd
Esc_dcd
          mov       ah,0E9h             ; <Esc>
          jmps      FF_dcd
Tab_dcd
          mov       ah,0E8h             ; <Tab>

     ;[ al <-- FFh ]
FF_dcd
          mov       al,0FFh             ; for special key, or extended kb code

          ;--------------------
          stosw                         ; write  FFh + scan or special code
          mov       pMacMark,di         ; store position in MACBUFF
          ret

     ;[ copy string to MACBUFF ]
Str_dcd
          mov       cx,ax               ; length of string
          mov       si,SYMBUFF
          rep
          movsb
          and       Flag, not STRING    ; clear flag
          mov       pMacMark,di         ; store position in MACBUFF
          ret
          ENDP
; WRITEBUFF 
;͸
;             WriteBuff                
;   writes buffer to file or STDOUT    
;                                      
; In: di = limit of output data        
;     MACINPUT flag set if input file  
;       is macro file                  
;;
WriteBuff PROC      NEAR
          test      Flag,NOWRITE        ; was error found in text file?
          jnz       End_wrb             ;    yes, done
          call      GetArg              ; second argument on command line?
          jnc       Cre_wrb             ;    yes, si = ptr to filename. jump
                                        ;    no, output must be text
;=======================================
;[ I. write TEXTBUFF to STDOUT ]
          mov       bp,di               ; bp = limit of data
          mov       di,TEXTBUFF         ; di = starting point
          sub       bp,di               ; bp = limit - start = bytes remaining

     ;[ screen loop ]
NxtS_wrb
          mov       cx,bp               ; cx = bytes remaining
          mov       dx,di               ; save start in dx = ptr to buffer
          mov       ah,Rows             ; ah = rows/screen - 1
          dec       ah                  ; ah = max lines to write
          mov       al,LF

     ;[ line loop ]
NxtL_wrb
          repne
          scasb
          jcxz      Wri_wrb             ; if end of data, jump to write
          dec       ah                  ; count one line
          jnz       NxtL_wrb            ; if more lines to count, jump

Wri_wrb
          mov       bp,cx               ; save bytes remaining
          mov       cx,di               ; di = limit
          sub       cx,dx               ; cx = limit - start = bytes to write
          call      WriteOut
          or        bp,bp               ; any bytes left?
          jz        End_wrb             ;   no, done
          call      Pause
          jmps      NxtS_wrb            ;   yes, jump to top of screen loop
End_wrb
          ret

;=======================================
;[ II. write to file ]

     ;[ create file ]
Cre_wrb
          mov       dx,si               ; dx = ptr to filename
          xor       cx,cx               ; cx = normal attribute
          mov       ah,03Ch             ; create file
          int       21h
          jc        Xc_wrb              ; jump on error
          mov       bx,ax               ; bx = handle

     ;[ set dx (ptr to buffer) ]
          mov       dx,TEXTBUFF         ; initialize dx to TEXTBUFF
          test      Flag,MACINPUT       ; was input file a MACRO file?
          jnz       DOS_wrb             ;   yes, OK. jump
          mov       dx,MACBUFF          ;   no, set dx to MACBUFF

     ;[ write file ]
DOS_wrb
          mov       cx,di               ; di = limit of output data
          sub       cx,dx               ; cx = limit - start = bytes to write
          mov       ah,040h             ; write file (bx = handle)
          int       21h
          pushf                         ; save flags
          mov       dx,ax               ; save bytes written in dx

     ;[ close file ]
          mov       ah,03Eh
          int       21h

     ;[ check for write error ]
          popf                          ; restore flags from write operation
          jc        Xw_wrb              ;   jump if problem
          cmp       cx,dx               ; requested bytes = bytes written?
          jnz       Xw_wrb              ;   no, error. jump

     ;[ write message ]
          mov       si,offset Cmsg      ; "completed" message
          call      PutMsg
          ret
;---------------------------------------
     ;[ error handling ]
Xc_wrb
          mov       si,offset Xcmsg     ; "can't create" message
          jmps      X_wrb
Xw_wrb
          mov       dx,si               ; dx = ptr to filename
          mov       ah,041h             ; delete file
          int       21h
          mov       si,offset Xwmsg     ; "write problem" message
X_wrb
          call      ErrorExit
          ENDP
; WRITING TO STDOUT 
;͸
;              LineError               
; writes current text line and error   
; message to STDOUT and sets flags     
;                                      
;  In: si = ptr to error message       
; Out: NOWRITE flag set                
;      CF set                          
;;
LineError PROC      NEAR
          call      PutLine
          call      PutMsg
          or        Flag,NOWRITE
          stc
          ret
          ENDP
;͸
;              SymError                
;writes current text line, contents of 
;SYMBUFF, and error message to STDOUT; 
;sets flags                            
;                                      
; In: si = ptr to message              
;Out: NOWRITE flag set                 
;     CF set                           
;;
SymError  PROC      NEAR
     ;[ write current line ]
          call      PutLine

     ;[ write indentation ]
          mov       dx,offset Indent    ; 4 spaces
          mov       cx,INDENTLEN
          call      WriteOut

     ;[ write symbol ]
          mov       dx,SYMBUFF          ; dx = ptr to string
          mov       di,dx               ; find length of string
          mov       cx,8                ; max. length of symbol is 7 bytes
          xor       al,al
          repne
          scasb
          neg       cx
          add       cx,7                ; cx = bytes to write
          call      WriteOut

     ;[ write error message ]
          mov       si,offset Xsmsg     ; "bad symbol" message
          test      Flag,KEY
          jz        P_syer
          mov       si,offset Xkmsg     ; "bad key" message
P_syer
          call      PutMsg

     ;[ set flags ]
          or        Flag,NOWRITE
          stc
          ret
          ENDP
;͸
;              PutLine                 
;  writes current text line to STDOUT  
;  pauses display when screen full     
;;
PutLine   PROC      NEAR
     ;[ pause if necessary ]
          sub       ErrCount,3
          ja        Go_pl               ; if ErrCount positive, no pause. jump
          mov       cl,Rows
          mov       ErrCount,cl         ; reset count
          call      Pause
Go_pl
     ;[ write CRLF ]
          mov       dx,offset Crlf
          mov       cx,2                ; cx = bytes to write
          call      WriteOut

     ;[ write line ]
          mov       dx,pLine            ; dx = ptr to line
          mov       cx,pLineMark        ; pLineMark = limit of line
          sub       cx,dx               ; cx = bytes to write
          call      WriteOut

     ;[ write CRLF ]
          mov       dx,offset Crlf
          mov       cx,2                ; cx = bytes to write
          call      WriteOut

          ret
          ENDP
;͸
;                PutMsg                
;  writes message and CRLF to STDOUT   
;                                      
; In: si = ptr to string with 1st word 
;          containing the length       
;;
PutMsg    PROC      NEAR
          lodsw
          mov       dx,si               ; dx = ptr to string
          mov       cx,ax               ; cx = bytes to write
          call      WriteOut
          mov       dx,offset Crlf
          mov       cx,2
          call      WriteOut
          ret
          ENDP
;͸
;               WriteOut               
;       writes string to STDOUT        
;                                      
; In:  dx = ptr to string              
;      cx = bytes to write             
;;
WriteOut  PROC      NEAR
          mov       bx,STDOUT           ; bx = device
          mov       ah,040h             ; write file or device
          int       21h
          jc        X_wro
          ret
;---------------------------------------
X_wro
          mov       si,offset Xpmsg
          mov       ah,0Eh              ; BIOS teletype function
Xx_wro
          lodsb
          or        al,al
          jz        Xz_wro
          int       10h
          jmps      Xx_wro
Xz_wro
          mov       ax,04CFFh
          int       21h
          ENDP
;͸
;                Pause                 
; pause output to STDOUT and display   
; "press a key" message                
;;
Pause     PROC      NEAR
          mov       dx,offset Pmsg      ; "press a key" message
          mov       cx,PLEN
          call      WriteOut
          xor       ax,ax               ; wait for keystroke
          int       16h
          mov       dx,offset Qmsg      ; erase message
          mov       cx,QLEN
          call      WriteOut
          ret
          ENDP
;͸
;              ErrorExit               
;  In: si = ptr to error message       
;;
ErrorExit PROC      NEAR
          call      PutMsg
          mov       ax,04CFFh
          int       21h
          ENDP
; DATA 

BytPerMac dw        ?                   ; bytes allotted for {key +} definition
ElemLen   dw        ?                   ; left element of table (BinSearch)
RemBytes  dw        ?                   ; bytes remaining in TEXTBUFF
SymbLen   dw        ?                   ; length of symbol (BinSearch)
pArg      dw        80h                 ; ptr to command line argument
pLine     dw        ?                   ; ptr to current line in TEXTBUFF
pLineMark dw        ?                   ; ptr marking limit of current line
pMacMark  dw        ?                   ; ptr marking postion in MACBUFF
pMacStart dw        ?                   ; ptr to start of macro in MACBUFF
pSymMark  dw        ?                   ; ptr marking limit of symbol in TEXTBUFF

Rows      db        ?
ErrCount  db        8                   ; counter for number of error messages
                                        ;   written to screen. used for pause.
Flag      db        0
MACINPUT  EQU      1b                   ; set if input file is a macro file
KEY       EQU     10b                   ; set if converting symbol for key
NOWRITE   EQU    100b                   ; set on error when making text file
STRING    EQU   1000b                   ; set if symbol in text file is a string
NEWKEY    EQU  10000b                   ; set if keys created by NewKey are OK.

STDOUT    EQU        1
LF        EQU       0Ah
CR        EQU       0Dh
EOF       EQU      01Ah
TRIGCODE  EQU      0E6h                 ; code for trigger key
EXTKBCODE EQU      0E0h                 ; code for extended keyboard keys

MAXSCAN   EQU       179                 ; number of elements in SpecSymb table

; STRINGS 

;                           ----- Usage Message -----
Umsg dw   ULEN
     db   '         DMC.COM (DDKEY Macro Compiler) (V. 1.0)         Richard Cowell',CR,LF
     db   '            Converts DDKEY macro files to text files and vice versa.',CR,LF
     db   '                   Usage:  DMC  {/N}  inputfile  {outputfile}',CR,LF
     db   '                   Option:  /N = use keys created by NewKey',CR,LF
     db   '         --------------------------------------------------------------',CR,LF
     db   'Format of Text File:',CR,LF
     db   'DDKEY               ;First line begins with ',27h,'DDKEY',27h,CR,LF
     db   'nn                  ;Second line begins with bytes/macro (5 - 96).',CR,LF
     db   'Key  =  Definition  ;Subsequent lines contain Key + = + Definition made up of',CR,LF
     db   ' .   .      .       ;valid symbols separated by spaces.',CR,LF
     db   CR,LF
     db   'Regarding symbols:',CR,LF
     db   '     @ means <Alt>      $ means <Shift>         c or C means Cursor Keypad',CR,LF
     db   '     ^ means <Ctrl>     * means Trigger Key     n or N means Numeric Keypad',CR,LF
     db   CR,LF
     db   '     Strings are delimited by ',27h,' or ".',CR,LF
     db   '     To enter either character as itself in a definition, use ',27h,27h,' or "".',CR,LF
     db   CR,LF
     db   '     <nnn> = ASCII, <snnn> = Scan, <ennn> = Extended code  (decimal)',CR,LF
     db   CR,LF
     db   'Blank lines and lines beginning with a space are ignored.',CR,LF
     db   'Comments at the beginning of a line or after a definition must begin with a',CR,LF
     db   'semicolon followed immediately by a non-space character, e.g. ;; ;T ;[.'
ULEN EQU  $ - offset Umsg - 2

;
;                   ----- Strings written to text file -----

Ddkey     db   'DDKEY   Macro File:  '
DDKEYLEN  EQU  $ - offset Ddkey

Nmsg      db   '  ==  Number of bytes per macro',CR,LF,CR,LF
NLEN      EQU  $ - offset Nmsg

;
;                      ----- Miscellaneous messages -----

Cmsg      db   CLEN,'     *=*=*=*=* Conversion complete *=*=*=*=*'
CLEN      EQU  $ - offset Cmsg - 2

Pmsg      db   '            (Press a key)    ',CR
PLEN      EQU  $ - offset Pmsg

Qmsg      db   '                         ',CR        ; for erasing Pmsg
QLEN      EQU  $ - offset Qmsg

Crlf      db   CR,LF,CR,LF

Indent    db   '    '
INDENTLEN EQU  $ - offset Indent

;
;                          ----- Error Messages -----

Xamsg     db   XALEN,'    This macro key has already been used.'
XALEN     EQU  $ - offset Xamsg - 2

Xbmsg     db   XBLEN,CR,LF,'The number of bytes per macro must be between 5 and 96 inclusive.'
XBLEN     EQU  $ - offset Xbmsg - 2

Xcmsg     db   XCLEN,CR,LF,'Cannot create output file.'
XCLEN     EQU  $ - offset Xcmsg - 2

Xdmsg     db   XDLEN,'    There is no definition for this macro key.'
XDLEN     EQU  $ - offset Xdmsg - 2

Xemsg     db   XELEN,'    There is a problem with the equal sign.'
XELEN     EQU  $ - offset Xemsg - 2

Xfmsg     db   XFLEN,CR,LF,'The file is too big to handle.'
XFLEN     EQU  $ - offset Xfmsg - 2

Ximsg     db   XILEN,CR,LF,'The input file is neither a DDKEY macro file nor a DMC text file.'
XILEN     EQU  $ - offset Ximsg - 2

Xkmsg     db   XKLEN,'  <--  This is not a valid symbol for a macro key.'
XKLEN     EQU  $ - offset Xkmsg - 2

Xlmsg     db   XLLEN,CR,LF,'The text file being made is too long for the available space (45,000 bytes).'
XLLEN     EQU  $ - offset Xlmsg - 2

Xmmsg     db   XMLEN,CR,LF,'There are too many macros in the text file.'
XMLEN     EQU  $ - offset Xmmsg - 2

Xnmsg     db   XNLEN,CR,LF,'An output filename is needed to convert a text file to a macro file.'
XNLEN     EQU  $ - offset Xnmsg - 2

Xomsg     db   XOLEN,CR,LF,'Problem opening input file.'
XOLEN     EQU  $ - offset Xomsg - 2

Xpmsg     db   CR,LF,'Problem writing to screen with DOS function.',0

Xqmsg     db   XQLEN,'    There is no matching quotation mark at the end of the string in this line.'
XQLEN     EQU  $ - offset Xqmsg - 2

Xrmsg     db   XRLEN,CR,LF,'Problem reading input file.'
XRLEN     EQU  $ - offset Xrmsg - 2

Xsmsg     db   XSLEN,'  <--  This is not a valid symbol.'
XSLEN     EQU  $ - offset Xsmsg - 2

Xtmsg     db   XTLEN,'    This line contains too many bytes.'
XTLEN     EQU  $ - offset Xtmsg - 2

Xwmsg     db   XWLEN,CR,LF,'Problem writing output data.'
XWLEN     EQU  $ - offset Xwmsg - 2

; INCLUDE 

          include 'DMCTABLE'

; HEAP 
STACKSIZE EQU  64
STACKTOP  EQU  ($+STACKSIZE) AND 0FFFEh ; end of code + stack size
                                        ;   rounded off to even number
     ;[ symbol buffers ]
SYMBUFF   EQU  STACKTOP + 2             ; allow 2-byte margin
SYMSIZE   EQU  64                       ; allow room for strings
BSBUFF    EQU  SYMBUFF + SYMSIZE        ; BSBUFF is used by BinSearch procedure

     ;[ macro buffer ]
MACBUFF   EQU  BSBUFF + SYMSIZE         ; start of MACBUFF = limit of BSBUFF
MACSIZE   EQU  12964                    ; largest DDKEY buffer is 12,960 bytes.
                                        ;    allow 2 bytes for bytes/macro and
                                        ;    2 bytes for terminal 0000h
MBUFFEND  EQU  MACBUFF + MACSIZE - 2    ; last byte of MACBUFF

     ;[ text buffer ]
TEXTBUFF  EQU  MACBUFF + MACSIZE        ; start of TEXTBUFF = limit of MACBUFF
HEAPEND   EQU  0FFFFh                   ; end of code/data/stack segment
TEXTSIZE  EQU  HEAPEND - TEXTBUFF       ; size of TEXTBUFF (about 46,000 bytes)
TBUFFLIM  EQU  HEAPEND - 800            ; if the text file being made extends
                                        ;  to this point, it is much too long.
